'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

// Given a pipeline, collapses it to a filter, a window and an ordering
let compile = (() => {
  var _ref2 = _asyncToGenerator(function* (scope, pipeline) {
    const operations = pipeline.operations;
    if (operations.length == 0) {
      throw new Error("Empty pipelines can't be collapsed");
    }
    if (operations[0].op != 'source') {
      throw new Error("Can't collapse pipelines that does not start with an *");
    }

    const sourceId = operations[0].id;

    const fetchSpec = new FetchSpec(scope, sourceId);
    operations.slice(1).reverse().forEach(function (operation) {
      (0, _debug2.default)('compiling', operation);
      switch (operation.op) {
        case 'filter':
          fetchSpec.applyFilter(operation.filter);
          break;
        case 'object':
          fetchSpec.project(operation.operations);
          break;
        case 'subscript':
          fetchSpec.applyWindow(operation.start, operation.end);
          break;
        case 'ordering':
          fetchSpec.applyOrdering(operation.terms);
        case 'accessor':
          fetchSpec.project(operation.operations);
          break;
        default:
          throw new Error(`Unknown pipeline operation ${operation.op} when collapsing fetche operation`);
      }
    });

    const generalizedFilter = (0, _generalizeJoinFilter2.default)(fetchSpec.filter, scope);
    const wasJoin = generalizedFilter !== fetchSpec.filter;

    // If this was a join, replace fetch with a join-fetch
    if (wasJoin) {
      const joinFetch = new FetchSpec(scope, sourceId);
      joinFetch.filter = generalizedFilter;
      return joinFetch;
    }

    // TODO: Now the filter needs to be expanded in case it requires anything
    // to be fetched in order to be executed.
    return fetchSpec;
  });

  return function compile(_x4, _x5) {
    return _ref2.apply(this, arguments);
  };
})();

var _operations = require('../plan/operations');

var _Scope = require('./Scope');

var _Scope2 = _interopRequireDefault(_Scope);

var _debug = require('../debug');

var _debug2 = _interopRequireDefault(_debug);

var _generalizeJoinFilter = require('./generalizeJoinFilter');

var _generalizeJoinFilter2 = _interopRequireDefault(_generalizeJoinFilter);

var _path = require('path');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

// Given a pipeline, fetch will make sure the source value for that pipeline
// gets cached in the parentScope. It does not actually return the retrieved
// value, but the same value with the fetch result cached into the sourceId
// of the provided pipeline.
// The fetcher is expected to return an object that looks something like this
// {
//   results: [{_id: 'doc4', _type: 'document'}, {_id: 'doc5', _type: 'document'}, ref: {_ref: 'doc4', weak: true}],
//   start: 3,
//   refs: {'doc5': [{id: 'doc3', weak: true}]}
// }
// `results` is required, `start` is used to signal that this set is allready offset by the fetcher so that the
// execution engine will not double-offset it, `refs` is a flattened list of all refs in every included document.
// This is optional and will be generated on demand if missing.

exports.default = (() => {
  var _ref = _asyncToGenerator(function* (scope, pipeline, fetcher) {
    (0, _debug2.default)('fetch()', scope, pipeline);
    const sourceId = extractSourceId(pipeline);
    // Is it a fetchable source, and did we not fetch it allready?
    if (!sourceId || scope.sourceIsCached(sourceId)) {
      (0, _debug2.default)('fetch() short circuit', sourceId, scope.sourceIsCached(sourceId));
      return scope;
    }
    // We need to actually go get it
    const fetchSpec = yield compile(scope, pipeline);
    const response = yield fetcher(fetchSpec);
    // Cache the result on the scope so it will bind to the source
    // when it comes up later
    (0, _debug2.default)('fetch() =>', response);
    scope.cacheRefs(response.refs);
    return scope.cacheSource(sourceId, {
      documents: response.results,
      start: response.start || 0
    });
  });

  function fetch(_x, _x2, _x3) {
    return _ref.apply(this, arguments);
  }

  return fetch;
})();

// Keeps track of the central parameters of a fetch operation.
// Since pipelines are computed in reverse order, filters,
// orderings and windows are expected to be applied in reverse order
// last to first.


class FetchSpec {
  constructor(parentScope, sourceId) {
    this.parentScope = parentScope;
    this.sourceId = sourceId;
    this.clear();
  }

  applyFilter(constraint) {
    if (!this.filter) {
      this.filter = constraint;
      return;
    }
    this.filter = new _operations.BinaryOperator('and', this.filter, constraint);
  }

  project(objectOperations) {
    // TODO: Remap the values of the filter and orderings collected so far
    // by substituting expressions according to the projection
  }

  applyWindow(start, end) {
    (0, _debug2.default)("applyWindow()", start, end);
    if (!Number.isFinite(start) || !Number.isFinite(end)) {
      throw new Error("start and end must be finite when applying window");
    }
    if (this.start === null) {
      this.start = start;
      this.end = end;
      return;
    }
    this.start += start;
    this.end = Math.min(this.end + start, end);
  }

  applyOrdering(terms) {
    this.ordering = terms.concat(this.ordering);
  }

  clear() {
    this.filter = null;
    this.start = null;
    this.end = null;
    this.ordering = [];
  }
}

function extractSourceId(pipeline) {
  const operations = pipeline.operations;
  if (operations.length == 0) {
    (0, _debug2.default)('extractSourceId: empty pipeline');
    return null;
  }
  if (operations[0].op != 'source') {
    (0, _debug2.default)('extractSourceId: op 0 is not source', operations[0]);
    return null;
  }
  return operations[0].id;
}
//# sourceMappingURL=fetch.js.map